# Biofeedback Blasters
# a space-invaders clone that monitors your heart rate
# using special hardware, and pauses the game if your heart rate
# gets too high

# code developed by oldgrover, ne0nra1n and psychedelicbike
# copyright 2008
# this program is offered free of charge but not free of bugs,
# and is distributed under the GPLv3 license.

import pygame
from pygame.locals import *
import sys
import random
import time, wave, pymedia.audio.sound as sound

# the number of heart rate samples to collect and average
SAMPLE_NUM=10
# four possible states for the game to be in
INIT=0
PAUSED=1
CONFIG=2
PLAYING=3
# the size of the buffer that stores the signal from the heart rate
# monitor.  It might be necessary to increase this for very slow
# heart rates
BUF_SIZE=120
# pymedia only provides information from the input device in
# big chunks.  These chunks have to be split up into smaller
# chunks before being averaged in order to avoid too much low-pass
# filter effect.  Too much breaking up increases the processor load
# overmuch, and too little breaking up flattens the heartbeat signal.
# trial and error showed that 10 was a good number.
SAMPLE_BREAKDOWN=10

# calculate the average of several samples of heart rate
def bpm_mean(samplearray):
    sum=0
    for i in range(0,SAMPLE_NUM):
        sum += samplearray[i]
    return sum / SAMPLE_NUM

# calculate the standard deviation of several samples of heart rate
def bpm_stddev(samplearray,mean):
    sum=0
    for i in range(0,SAMPLE_NUM):
        sum += (samplearray[i]-mean)**2
    return (sum/SAMPLE_NUM)**0.5

# this function is used to check if the heart rate sample is within one
# standard deviation of the average of the past number of samples.
# Occasionally, the program misses a heartbeat, or is triggered by noise
# so this prevents capturing obviously incorrect data.
def bpm_check(sample, samplearray):
    mean = bpm_mean(samplearray)
    stddev = bpm_stddev(samplearray,mean)
    if (sample > mean + stddev) or (sample < mean - stddev):
        return False
    return True

# the heartbeat sound
soundFile = wave.open( 'heartbeat.wav', 'rb' )

# setting up pymedia for output
sampleRate= soundFile.getframerate()
channels= soundFile.getnchannels()
format= sound.AFMT_S16_LE

framesCount = soundFile.getnframes()

print "Sample Rate : " + str(sampleRate) + "  " + str(channels)  + " " + str(framesCount)
sndOutput= sound.Output( sampleRate, channels, format )
beepSound = soundFile.readframes(framesCount)

soundFile = wave.open( 'explosion-01.wav', 'rb' )
framesCount = soundFile.getnframes()
explosionsound = soundFile.readframes(framesCount)
sampleRate= soundFile.getframerate()
sndOutput2 = sound.Output( sampleRate, channels, format )

soundFile = wave.open( 'bloop1.wav', 'rb' )
framesCount = soundFile.getnframes()
bloopsound = soundFile.readframes(framesCount)

soundFile = wave.open('laser.wav', 'rb' )
framesCount = soundFile.getnframes()
lasersound = soundFile.readframes(framesCount)

soundFile = wave.open('woopsound.wav', 'rb' )
framesCount = soundFile.getnframes()
woopsound = soundFile.readframes(framesCount)
sampleRate= soundFile.getframerate()
sndOutput3 = sound.Output( sampleRate, channels, format )

soundFile = wave.open('topdudedeath.wav', 'rb' )
framesCount = soundFile.getnframes()
topdudedeathsound = soundFile.readframes(framesCount)
sampleRate= soundFile.getframerate()
sndOutput4 = sound.Output( sampleRate, channels, format )

soundFile = wave.open('playerdeath.wav', 'rb')
framesCount = soundFile.getnframes()
playerdeathsound = soundFile.readframes(framesCount)
sampleRate = soundFile.getframerate()
sndOutput5 = sound.Output(sampleRate, channels, format)

# setting up the sound input.  AFMT_S8 is IMPORTANT
# it means that the data coming in is in 8-bit data format, so it
# can be treated as chars (it is delivered as a string)
sndRecorder = sound.Input( 44100, 1, sound.AFMT_S8)
sndRecorder.start()

# the player-controlled character sprite
class Player(pygame.sprite.Sprite) :
    # most of the member variables are self explanatory
    def __init__(self,image,x,y,num):
        pygame.sprite.Sprite.__init__(self)
        self.image = image
        self.rect = self.image.get_rect()
        self.rect.topleft = [x,y]
        self.dx = 0
        self.dy = 0
        self.identity = num
        self.next_update_time = 0
        self.shooting = False
        self.next_shoot_time = 0
        self.playerpoints = 0
        self.playerlives = 3
        self.playerlevel = 1
        self.heartrate = 0
        self.averageheartrate = 0
        self.state = INIT

    # the game state is stored in the player sprite object
    # (hey, it's instantiated as a global.  It's just easier.)
    def setstate(self, state):
        self.state = state
        
    # used to access the player instance to find out what the current
    # state is
    def getstate(self):
        return self.state

    # set the target rate that the heart rate cannot exceed without
    # pausing the game
    def setaverageheartrate(self, averageheartrate) :
        self.averageheartrate = averageheartrate
        avgheartratefontsurface = pointfont.render(str(int(averageheartrate*1.2)),True,[255,255,255])
        screen.blit(pygame.Surface([120,40]), [640,380])
        screen.blit(avgheartratefontsurface, [640,380])
        updaterect = pygame.Rect(640,380,640+BUF_SIZE,420)
        pygame.display.update(updaterect)        

    # access the target heart rate
    def getaverageheartrate(self) :
        return self.averageheartrate

    # set the current heart rate, and update the display
    def setheartrate(self, heartrate) :
        if self.state == INIT:
            self.state = CONFIG
        self.heartrate = heartrate
        heartratefontsurface = pointfont.render(str(int(heartrate)),True,[255,255,255])
        screen.blit(pygame.Surface([120,40]), [640,300])
        screen.blit(heartratefontsurface, [640,300])
        updaterect = pygame.Rect(640,300,640+BUF_SIZE,340)
        pygame.display.update(updaterect)
        
    # access the current heart rate
    def getheartrate(self) :
        return self.heartrate

    # increase the level (called when the player destroys all the invaders)
    def increaselevel(self) :
        self.playerlevel = self.playerlevel + 1

    # access the current level that the player is on
    def getlevel(self) :
        return self.playerlevel

    # make sure the player does not drive off the screen
    def update(self,current_time):
        if self.next_update_time < current_time :
            self.rect.left = self.rect.left + self.dx
            if (self.rect.right > 640) :
                self.rect.right = 640
            if (self.rect.left < 1) :
                self.rect.left = 1
            self.next_update_time = current_time + 1

    # when the player gets shot, this is used to decrement their life count
    def removelife(self) :
        self.playerlives = self.playerlives - 1

    # access the number of lives the player has
    def getlives(self) :
        return self.playerlives

    # increment the number of lives the player has (called when level
    # advances)
    def addlives(self,lives) :
        self.playerlives = self.playerlives + lives

    # the method to increment the number of points the player has
    def addpoints(self,points) :
        self.playerpoints = self.playerpoints + points

    # access the number of points the player has
    def getpoints(self) :
        return self.playerpoints

    # access the x position of the player sprite
    def getx(self) :
        return self.rect.left

    # access the y position of the player sprite
    def gety(self) :
        return self.rect.top

    # this function allows for asynchronous firing control to be
    # triggered during the normal update cycle
    def startshooting(self) :
        if (self.next_shoot_time <  pygame.time.get_ticks()) :
            self.shooting = True
            self.next_shoot_time = pygame.time.get_ticks() + 400

    # yeah.  More of that synchronizing user input to the update cycle
    def stopshooting(self) :
        self.shooting = False

    # second comment, same as the first.  Little bit louder and
    # a little bit worse.
    def isshooting(self) :
        return self.shooting

    # stop the player sprite from moving
    def stop(self):
        self.dx = 0

    # get the player sprite moving to the left
    def goleft(self):
        self.dx = -4

    # get the player sprite moving to the right
    def goright(self):
        self.dx = 4
        
    # compares a shot sprite to the player sprite, and indicates
    # if the shot has hit the player
    def shothit(self,shot) :
        hasHit = False
        [x,y] = shot.getpos()
        if self.rect.collidepoint(x,y) :
            #kaboom!  you've been shot!  What should we do here?  Explosions, I guess
            hasHit = True
        return hasHit

# this class is for the laser beams shot by the player
class ShootingBeam(pygame.sprite.Sprite) :
    # most of the member variables are self-explanatory
    def __init__(self,image,x,y,num):
        pygame.sprite.Sprite.__init__(self)
        self.image = image
        self.rect = self.image.get_rect()
        self.rect.topleft = [x,y]
        self.dx = 0
        self.dy = -5
        self.identity = num
        self.next_update_time = 0

    # access the position of the laser beam
    def getpos(self) :
        return self.rect.topleft
        
    # move the shot up each update tick
    def update(self,current_time):
        if self.next_update_time < current_time :
            self.rect.top = self.rect.top + self.dy
            self.next_update_time = current_time + 1

# These are the laser beams shot by the invaders
class InvaderBeam(pygame.sprite.Sprite) :
    # most of the member variables are self-explanatory
    def __init__(self,image,x,y,num):
        pygame.sprite.Sprite.__init__(self)
        self.image = image
        self.rect = self.image.get_rect()
        self.rect.topleft = [x,y]
        self.dx = 0
        self.dy = 5
        self.identity = num
        self.next_update_time = 0

    # access the position of the laser beam
    def getpos(self) :
        return self.rect.topleft
        
    # move the laser beam down each update tick
    def update(self,current_time):
        if self.next_update_time < current_time :
            self.rect.top = self.rect.top + self.dy
            self.next_update_time = current_time + 1


# the player is currently controlling a 30x30 red square.  Yay.
playerimage = pygame.Surface([30,30])
playerimage.fill([255,0,0])
    
# instantiate the player sprite.  Hooray!
playersprite = Player(playerimage,(560/2)-15,440,0)

# pygame initialization stuff.
pygame.init()
screen = pygame.display.set_mode([640+BUF_SIZE,480])
background = pygame.Surface([640+BUF_SIZE,480])
background.fill([0,0,0])

pointsurface = pygame.Surface([75,40])
fontpath = pygame.font.match_font('arial', bold=False, italic=False)
pointfont = pygame.font.Font(fontpath,16)

pygame.display.set_caption("Biofeedback Blasters")

# A moving space invader
class Invader(pygame.sprite.Sprite):

    # I hope that most of the member variables here are self-explanatory
    def __init__(self,image,x,y,num):
        pygame.sprite.Sprite.__init__(self)
        self.image = image
        self.rect = self.image.get_rect()
        self.rect.topleft = [x,y]
        self.dx = 2
        self.dy = 0
        self.identity = num
        self.next_update_time = 0

    # need to be able to access the rectangles so we can check
    # for collisions
    def getrect(self) :
        return self.rect

    # we can increase the speed of the invaders as the level increases
    # as well as when they move down the screen
    def setspeed(self,speed) :
        if (self.dx > 0) :
            self.dx = speed
        else :
            self.dx= speed * -1

    # I don't think this is ever used.
    def getidentity(self) :
        return self.identity
    
    # move the space invader each update tick
    def update(self,current_time):
        if self.next_update_time < current_time :
            self.rect.left = self.rect.left + self.dx
            self.next_update_time = current_time + 4
        
    # adds the space invader to the invasion collection
    def joininvasion(self,invasion) :
        self.invasion = invasion
        invasion.add(self)

    # if an invader hits the edge of the screen, they all move down a bit
    def movedown(self) :
        self.rect.top = self.rect.top + 10

    # a check to see if the invader is at the edge of the screen
    def atedge(self) :
        if (self.rect.right >= 630) or (self.rect.left <= 10) :
            return True
        else :
            return False

    # when a space invader hits the edge of the screen, they all have
    # to reverse their direction (left->right, right->left)
    def flipdirection(self) :
        self.dx = self.dx * -1

    # access the y position of the invader
    def gety(self) :
        return self.rect.top

    # the invaders randomly shoot.  This is synchronized with the
    # update tick
    def isshooting(self) :
        #random number check to see if this invader is firing a shot
        shotPercentageInverse = 600
        if (random.randrange(shotPercentageInverse) == 0) :
            return True
        else :
            return False

# This is the dude that goes across the top of the screen, going
# "wub-wub-wub-wub" and is worth a lot of points
class TopDude(pygame.sprite.Sprite):
    # self-explanatory member variables
    def __init__(self,image,x,y,num):
        pygame.sprite.Sprite.__init__(self)
        self.image = image
        self.rect = self.image.get_rect()
        self.rect.topleft = [x,y]
        self.dx = 2
        self.dy = 0
        self.identity = num
        self.next_update_time = 0
        self.next_woop_time = 0

    # need to be able to access the rectangle the sprite is in
    # to be able to determine if there is a collision
    def getrect(self) :
        return self.rect

    # I'm still pretty sure this is never used.  It's the appendix of
    # the code.  It's like a vestigal tail, or maybe wisdom teeth.
    #  It may have served some purpose at some point in time, but
    # evolution has left this function behind.
    def getidentity(self) :
        return self.identity
    
    
    # each update tick the top dude moves towards the right
    def update(self,current_time):
        if self.next_update_time < current_time :
            self.rect.left = self.rect.left + self.dx
            self.next_update_time = current_time + 4
            if self.next_woop_time < current_time :
                sndOutput3.play(woopsound)
                self.next_woop_time = current_time + 1000
        
    # check to see if the top dude has reached the right edge of
    # the screen
    def atedge(self) :
        if (self.rect.right >= 630) :
            return True
        else :
            return False
    
# a collection of all individual invader instances
class Invasion :
    # I think the member variables are self-explanatory
    def __init__(self) :
        self.allInvaders = []
        self.hitwall = False
        self.answercount = 0
        self.speed=2
        self.bouncecounter = 0

    # the speed of the invaders is stored in the invasion
    def setspeed(self,speed) :
        self.speed = speed

    # the invasion is defeated when all the invaders are killed
    def invasiondefeated(self) :
        return len(self.allInvaders) == 0

    # given a player's laser beam, check each invader to determine
    # if the laser beam hits any of them
    def shothit(self,shot) :
        hasHit = False
        [x,y] = shot.getpos()
        for invader in self.allInvaders :
            if invader.getrect().collidepoint(x,y) :
                invader.kill()
                self.allInvaders.remove(invader)
                hasHit = True
        return hasHit
        
    # this does two things.  It checks if any invader has reached
    # the bottom (resulting in an end game).  It also checks to see
    # if any invader hits the left or right wall.  If it does, change the
    # direction of the invasion, and move all the invaders down
    def checkwallhits(self) :
        someonehitwall = False
        for invader in self.allInvaders :
            if invader.atedge() :
                someonehitwall = True
            if invader.gety() >= playersprite.gety() :
                print "Player should die"
                endgame()
        if someonehitwall :
            self.bouncecounter = self.bouncecounter + 1
            if self.bouncecounter == 3 :
                self.speed = self.speed + 1
                self.bouncecounter = 0
            for invader in self.allInvaders :
                invader.flipdirection()
                invader.movedown()
                invader.setspeed(self.speed)
    
    # add an invader to the collection
    def add(self,invader) :
        self.allInvaders.insert(0,invader)
        invader.setspeed(self.speed)

    # access the list of all invaders in the invasion
    def getinvaderlist(self) :
        return self.allInvaders
    
# keypress inputs
def input(events):
    for event in events:
        if event.type == QUIT:
            print event
            sys.exit(0)
        elif event.type == KEYDOWN:
            if event.key == K_ESCAPE :
                print event
                sys.exit(0)
            if event.key == K_LEFT :
                playersprite.goleft()
            if event.key == K_RIGHT :
                playersprite.goright()
            if event.key == K_SPACE :
                playersprite.startshooting()
            if event.key == K_p :
                playersprite.setstate(PAUSED)
        elif event.type == KEYUP:
            if (event.key == K_LEFT) or (event.key == K_RIGHT) :
                playersprite.stop()
            if (event.key == K_SPACE) :
                playersprite.stopshooting()
                
# the number of valid inputs is limited when the game is paused.
# the options are unpause and quit.
def pauseInput(events):
    # Here we check to see if we unpause the game.  You can still quit.
    for event in events:
        if event.type == QUIT:
            print event
            sys.exit(0)
        if event.type == KEYDOWN:
            if event.key == K_ESCAPE :
                print event
                sys.exit(0)
            if event.key == K_p:
                return False
    return True

# game over.
def endgame():
    global playersprite, allsprites, invasion, shotlist, invadershotlist, dudelist, invaderimage
    sndOutput5.play(playerdeathsound)
    exitsurface=pointfont.render("YOU'RE DEAD, SUCKER!",True,[255,255,255])
    screen.blit(exitsurface, [200,200])
    updaterect = pygame.Rect(200,200,300,230)
    pygame.display.update(updaterect)
    while True :
        for ev in pygame.event.get() :
            if ev.type == QUIT :
                sys.exit(0)
            if ev.type == KEYDOWN  :
                if (ev.key == K_n) :
                    sys.exit(0)
                if (ev.key == K_ESCAPE) :
                    sys.exit(0)
   
# when the player is shot, subtract a life, make sure they still have
# some lives left, and pause the game until the user presses a key
def killplayer():
    playersprite.removelife()
    sndOutput2.play(explosionsound)
    if (playersprite.getlives() == 0):
        endgame()
    keybeenpressed = False
    while not keybeenpressed :
        for ev in pygame.event.get() :
            if ev.type == QUIT :
                sys.exit(0)
            if ev.type == KEYDOWN  :
                keybeenpressed = True
    screen.blit(pointsurface, [400,30])
    lifefontsurface = pointfont.render(str(playersprite.getlives()), True, [255,255,255])
    screen.blit(lifefontsurface, [400,30])
    updaterect = pygame.Rect(400,30,499,60)
    pygame.display.update(updaterect)

# there are a limited number of valid inputs while trying to identify the
# heart rate.  You can wait, or you can quit.
def heartrateinput(events):
    for event in events:
        if event.type == QUIT:
            print event
            sys.exit(0)
        elif event.type == KEYDOWN:
            if event.key == K_ESCAPE :
                print event
                sys.exit(0)
        
# this is the heart of heart rate processing (hah!  I kill me!)        
def sampleProcess(samples):
    # okay, we created the heart rate sampler in a different program, and
    # moved it all into this one.  These global variables are a bubble gum
    # and duct tape solution to cobble it in here with the minimum
    # possible effort.
    global counter, totalcount, lastpos, trigger_level, buf_avg, buf_max, buffer, lastwasbeat, lasttick, sampleindex

    # the counter variable keeps track of where we are in the buffer
    counter = counter + 1
    # when the buffer is full, find out the maximum level, the average
    # and set the trigger level at half way between the max and the
    # average
    if counter == BUF_SIZE :
            counter = 0
            buf_avg = 0
            buf_max = 0
            for i in range(0,BUF_SIZE):
                buf_avg = buf_avg + buffer[i]
                if (buffer[i] > buf_max):
                    buf_max = buffer[i]
            buf_avg = buf_avg / BUF_SIZE
            trigger_level = (buf_avg + buf_max) / 2
    
    # here we take the average of all the values in the sample buffer
    tot = 0
    for sam in samples :
            tot = tot + ord(sam)
    avgval = (tot / len(samples))
    buffer[counter] = avgval
    totalcount = totalcount + 1
        
    # we can't find a heartbeat until we actually set the trigger level
    # hence this first if statement
    if totalcount > BUF_SIZE :
            # here's how we perform the rising edge trigger.
            if avgval >= trigger_level :
                    if not lastwasbeat :
                            # play a beep sound
                            if playersprite.getstate() == CONFIG or playersprite.getstate() == PAUSED:
                                sndOutput.play(beepSound)
                            # find out how long it's been since the last beat,
                            # calculate the heart rate, and check that it's valid
                            if lasttick != 0 :
                                diff = pygame.time.get_ticks() - lasttick
                                bps = 1000.0 / diff
                                bpm = bps * 60.0
                                sampleset[sampleindex] = bpm
                                sampleindex+=1
                                if sampleindex >= SAMPLE_NUM:
                                    sampleindex=0
                                if bpm_check(bpm,sampleset):
                                    #print str(bpm)
                                    playersprite.setheartrate(bpm)
                            lasttick = pygame.time.get_ticks()
                            lastwasbeat = True
            if avgval < trigger_level :
                    lastwasbeat = False
                    
    # these lines draw the little ekg monitor on the top right
    # of the screen
    pygame.draw.line(screen,[255,255,255],[counter+640,0],[counter+640,255])
    if counter != 0 :
            pygame.draw.line(screen,[255,0,0],lastpos,[counter+640,avgval])
    lastpos = [counter+640,avgval]

    if totalcount > BUF_SIZE :
       pygame.draw.line(screen,[0,255,0],[640,trigger_level],[640+BUF_SIZE,trigger_level])
       pygame.draw.line(screen,[0,0,255],[640,buf_avg],[640+BUF_SIZE,buf_avg])
    updaterect = pygame.Rect(640,0,640+BUF_SIZE,255)
    pygame.display.update(updaterect)

# a whole bunch of ugly kludgy global variables
counter = 0
buffer = []
totalcount = 0
lastpos = [640,0]

overtot = 0
oversample = 0
maxsample = 0
trigger_level = 0
buf_avg = 0
buf_max = 0

variance = 0

lastwasbeat = False

lasttick = 0

sampleindex=0
sampleset = []

shotlist = []
invadershotlist = []

invasion = Invasion()
dudelist = []

invaderimage = []

# um, this is, like, the main function for the program.  My dog is
# now chasing a bug across my living room.  He wants to eat it.
def main():
    # a bunch of configuration stuff.  Not that exciting.
    configcounter = 0
    averageheartrate = 0
    initcounter = 0
    for i in range(0,SAMPLE_NUM):
        sampleset.append(75)
    for i in range(0,BUF_SIZE):
        buffer.append(0)

    # pygame.draw.line(screen,[255,255,255],[561,0],[561,480])
    pygame.display.flip()

    image = pygame.image.load("red_grand.jpg")
    invaderimage.append(image)
    image = pygame.image.load("blue_grand.jpg")
    invaderimage.append(image)
    image = pygame.image.load("green_grand.jpg")
    invaderimage.append(image)
    image = pygame.image.load("yellow_grand.jpg")
    invaderimage.append(image)
    
    allsprites = pygame.sprite.RenderUpdates() # specialtype of sprite group

    # create all the invaders
    for y in xrange(0,4) :
        for x in xrange(0,7) :
            invader = Invader(invaderimage[y],((x * 60)+30),(y * 40)+100,(y*7)+x);
            invader.joininvasion(invasion)
            allsprites.add(invader)
    
    allsprites.add(playersprite)

    # more initialization stuff
    beamimage = pygame.Surface([1,10])
    beamimage.fill([0,255,0])

    invaderbeamimage = pygame.Surface([1,10])
    invaderbeamimage.fill([0,255,255])

    scorefontsurface = pointfont.render("SCORE",True,[255,255,255])
    screen.blit(scorefontsurface,[30,30])
    pointfontsurface = pointfont.render(str(playersprite.getpoints()),True,[255,255,255])
    screen.blit(pointfontsurface,[100,30])
    lifetextfontsurface = pointfont.render("LIVES",True,[255,255,255])
    screen.blit(lifetextfontsurface,[300,30])
    lifefontsurface = pointfont.render(str(playersprite.getlives()),True,[255,255,255])
    screen.blit(lifefontsurface, [400,30])
    leveltextfontsurface = pointfont.render("LEVEL",True,[255,255,255])
    screen.blit(leveltextfontsurface,[500,30])
    levelfontsurface = pointfont.render(str(playersprite.getlevel()),True,[255,255,255])
    screen.blit(levelfontsurface, [580,30])
    heartratetextfontsurface = pointfont.render("HEARTRATE",True,[255,255,255])
    screen.blit(heartratetextfontsurface,[640,260])
    limittextfontsurface = pointfont.render("LIMIT",True,[255,255,255])
    screen.blit(limittextfontsurface,[640,340])

    pygame.display.flip()

    topDudeCounter = 0
    topDudeNumber = 100
    topDudeImage = pygame.image.load("defcon_ship.jpg")
        
    # the main loop of the program
    while True :
        # a delay so the game doesn't run at lightning speed
        pygame.time.delay(20)

         # grab and process the data from the sound input device
        samples = sndRecorder.getData()
        if samples and len( samples ):
            #break up the "samples" string according to the
            #constant "SAMPLE_BREAKDOWN"
            chunksize = len(samples)/SAMPLE_BREAKDOWN
            for i in range(0,SAMPLE_BREAKDOWN):
                sampleProcess(samples[i*chunksize:(i+1)*chunksize-1])

        # here's the code that runs when the game is running
        if playersprite.getstate() == PLAYING:
            # automatic pause if the heart rate is too high
            if playersprite.getheartrate() >= playersprite.getaverageheartrate() * 1.2:
                playersprite.setstate(PAUSED)
                
            # update the display of all the sprites
            allsprites.update(pygame.time.get_ticks())
            rectlist = allsprites.draw(screen)
            pygame.display.update(rectlist)

            # every once in a while, release a new top dude
            topDudeCounter += 1
            if topDudeCounter >= 400:
                topDudeCounter = 0
                topDude = TopDude(topDudeImage,30,60,topDudeNumber)
                topDudeNumber += 1
                dudelist.insert(0,topDude)
                allsprites.add(topDude)
            allsprites.clear(screen,background)

            # here we check if any of the invaders have hit a wall
            invasion.checkwallhits()

            # grab any inputs
            input(pygame.event.get())

            # create a laser beam from the player if he's shooting
            if (playersprite.isshooting()) :
                shot = ShootingBeam(beamimage,playersprite.getx()+15,405,0)
                sndOutput2.play(lasersound)
                allsprites.add(shot)
                shotlist.insert(0,shot)
                playersprite.stopshooting()

            # get a list of all active invaders
            activeInvaderList = invasion.getinvaderlist()
            
            # for each invader in the list, check if it is shooting (randomly)
            # if the invader is shooting, create a new invaderShot, and add it
            # to the invadershotlist
            for activeInvader in activeInvaderList:
                if (activeInvader.isshooting()):
                    invaderRect = activeInvader.getrect()
                    invaderXpos = invaderRect.centerx
                    invaderYpos = invaderRect.bottom
                    shot = InvaderBeam(invaderbeamimage, invaderXpos, invaderYpos, 0)
                    allsprites.add(shot)
                    invadershotlist.insert(0,shot)

            # look to see if the player gets shot (eventually)
            playerwasshot = False
            for shot in invadershotlist :
                if playersprite.shothit(shot) :
                    shot.kill()
                    invadershotlist.remove(shot)
                    killplayer()
                    playerwasshot = True
                [shotx,shoty] = shot.getpos()
                if shoty > 480 :
                    shot.kill()
                    invadershotlist.remove(shot)

            if playerwasshot :
                for shot in invadershotlist :
                    shot.kill()
                    invadershotlist.remove(shot)

            # look to see if any invader gets shot
            for shot in shotlist :
                if invasion.shothit(shot) :
                    shot.kill()
                    shotlist.remove(shot)
                    playersprite.addpoints(10)
                    sndOutput2.play(bloopsound)
                    screen.blit(pointsurface,[100,30])
                    pointfontsurface = pointfont.render(str(playersprite.getpoints()),True,[255,255,255])
                    screen.blit(pointfontsurface,[100,30])
                    updaterect = pygame.Rect(100,30,175,60)
                    pygame.display.update(updaterect)
                [shotx,shoty] = shot.getpos()
                if shoty < 0 :
                    shot.kill()
                    shotlist.remove(shot)

            # look to see if any topDude gets shot
            for shot in shotlist :
                for topDude in dudelist:
                    if topDude.getrect().collidepoint(shot.getpos()) :
                        topDude.kill()
                        dudelist.remove(topDude)
                        shot.kill()
                        shotlist.remove(shot)
                        playersprite.addpoints(50)
                        sndOutput4.play(topdudedeathsound)
                        screen.blit(pointsurface,[100,30])
                        pointfontsurface = pointfont.render(str(playersprite.getpoints()),True,[255,255,255])
                        screen.blit(pointfontsurface,[100,30])
                        updaterect = pygame.Rect(100,30,175,60)
                        pygame.display.update(updaterect)

            # Check to see if any topdudes have gone off the edge of the screen,
            # never to be seen again
            for topDude in dudelist :
                if topDude.atedge() :
                    topDude.kill()
                    dudelist.remove(topDude)
            
            # check to see if the player has managed to kill all invaders
            if invasion.invasiondefeated() :
                playersprite.addpoints(100)
                playersprite.addlives(1)
                playersprite.increaselevel()
                invasion.setspeed(1+playersprite.getlevel())
                screen.blit(pointsurface,[100,30])
                pointfontsurface = pointfont.render(str(playersprite.getpoints()),True,[255,255,255])
                screen.blit(pointfontsurface,[100,30])
                updaterect = pygame.Rect(100,30,175,60)
                pygame.display.update(updaterect)
                screen.blit(pointsurface, [580,30])
                lifefontsurface = pointfont.render(str(playersprite.getlevel()), True, [255,255,255])
                screen.blit(lifefontsurface, [580,30])
                updaterect = pygame.Rect(580,30,640,60)
                pygame.display.update(updaterect)
                screen.blit(pointsurface, [400,30])
                lifefontsurface = pointfont.render(str(playersprite.getlives()), True, [255,255,255])
                screen.blit(lifefontsurface, [400,30])
                updaterect = pygame.Rect(400,30,499,60)
                pygame.display.update(updaterect)

                # remove all shots that may be remaining
                for shot in invadershotlist :
                    shot.kill()
                    invadershotlist.remove(shot)
                for shot in shotlist :
                    shot.kill
                    shotlist.remove(shot)

                # create a new wave of invaders to shoot at
                for y in xrange(0,4) :
                    for x in xrange(0,7) :
                        invader = Invader(invaderimage[y],((x * 60)+30),(y * 40)+100,(y*7)+x);
                        invader.joininvasion(invasion)
                        allsprites.add(invader)

        # This is where the heart rate is initializing
        if playersprite.getstate() == INIT:
            initcounter += 1
            if initcounter >= 25:
                #print "Trying to find your heartrate"
                initcounter = 0
            heartratefontsurface = pointfont.render("FINDING",True,[255,255,255])
            screen.blit(pygame.Surface([120,40]), [640,300])
            screen.blit(heartratefontsurface, [640,300])
            updaterect = pygame.Rect(640,300,640+BUF_SIZE,340)
            pygame.display.update(updaterect)
            heartrateinput(pygame.event.get())
            
        # this is where we're trying to grab your average initial heartrate
        if playersprite.getstate() == CONFIG:
            avgheartratefontsurface = pointfont.render("CALCULATING",True,[255,255,255])
            screen.blit(pygame.Surface([120,40]), [640,380])
            screen.blit(avgheartratefontsurface, [640,380])
            updaterect = pygame.Rect(640,380,640+BUF_SIZE,420)
            pygame.display.update(updaterect)
            heartrateinput(pygame.event.get())
            initcounter += 1
            if initcounter >= 100:
                initcounter = 0
                #print "Please wait, calibrating average heart rate"
                #print "Collecting sample ",configcounter,"heartrate is ",playersprite.getheartrate()
                averageheartrate += playersprite.getheartrate()/10.0
                configcounter += 1
                if configcounter >= 10:
                    playersprite.setstate(PLAYING)
                    print "Average heart rate is",averageheartrate
                    playersprite.setaverageheartrate(averageheartrate)
                    
        # this is what runs when the game is paused
        if playersprite.getstate() == PAUSED:
            paused = True
            paused = pauseInput(pygame.event.get())
            # we only allow you to unpause if you've calmed down
            # and your heart rate is below the limit again
            if not(paused) and playersprite.getheartrate() >= playersprite.getaverageheartrate() * 1.2:
                print "Heartrate too high to unpause"
                paused = True
            if not(paused):
                playersprite.setstate(PLAYING)
                
# kick it off!
main()
